/*
 *	DATA.C
 *	Tom Kerrigan's Simple Chess Program (TSCP)
 *
 *	Copyright 1997 Tom Kerrigan
 *
 *   	Traduzione in italiano di Raffaele Carnuccio
 */


#include "defs.h"


/*la rappresentazione della scacchiera */
int color[64];  /* LIGHT, DARK, or EMPTY : bianco, nero o vuoto*/
int piece[64];  /* PAWN, KNIGHT, BISHOP, ROOK, QUEEN, KING, or EMPTY :
		   pedone, cavallo, alfiere, torre, donna, re o vuoto*/
int side;       /* il lato che ha il tratto */
int xside;      /* il lato che non ha il tratto*/
int castle;     /* un campo di bit che determina i permessi per gli arrocchi:
			0001	1	-->	il bianco pu arroccare corto
			0010	2	-->	il bianco pu arrocare lungo
			0100	4	-->	il nero pu arroccare corto
			1000	8	-->	il nero pu arroccare lungo
		La combinazione di 1 e 0 determina i permessi per ambo i colori:
		Esempio:
		int castle = 5 --> 5 in binario  0111 --> il bianco pu arroccare sia corto
		che lungo, mentre il nero pu arroccare solo corto (infatti il byte dell'arrocco
		lungo, il quarto da destra,  settato a 0). [N.d.T]*/

int ep;	      /* la casa en-passant. Se il bianco muove e2e4, la casa en-passant  settata ad e3, 
            	 perch  l'unica casa dove un pedone nero pu catturare en-passant */

int fifty;    /* il numero di mosse trascorse dall'ultima cattura o dall'ultima spinta di un pedone;
		  utilizzata per verificare una patta per la regola delle 50 mosse*/

int ply;      /*  il numero di semimosse (ply) calcolate dalla posizione da cui la ricerca  partita
		 in altri termini  la profondit raggiunta dell'albero a partire dalla sua radice (root) */

int hply;     /* h sta per history (storia);  il numero di semimosse giocate dall'inizio della partita */


/* gen_dat  una zona di memoria riservata alla lista delle mosse create per il 
   generatore di mosse.
   La lista di mosse per la profondit n inizia con first_move[n] e finisce con
   first_move[n + 1]. */

gen_t gen_dat[GEN_STACK];
int first_move[MAX_PLY];

/* il vettore per la storia euristica (usata per l'ordinamento di mosse) */
int history[64][64];

/* abbiamo bisogno di un vettore di hist_t per poter tornare indietro di una o pi mosse */
hist_t hist_dat[HIST_STACK];

/* l'engine cercher la migliore mossa per max_time millisecondi oppure fino a quando non sia
   stata raggiunta la profondit max_depth. */
int max_time;
int max_depth;

/* il momento in cui l'engine inizia a calcolare, e quello in cui dovrebbe fermarsi */
int start_time;
int stop_time;

int nodes;  /* il numero di nodi visitati */

/* un vettore triangolare per la gestione della PV (Variante Principale) */
move pv[MAX_PLY][MAX_PLY];
int pv_length[MAX_PLY];
BOOL follow_pv;


/* Ora abbiamo il vettore mailbox, cos chiamato perch esso ricorda
   una cassetta di posta, almeno secondo l'opinione di Bob Hyatt.
   Questo ci  utile quando abbiamo bisogno di capire dove pu andare un
   certo pezzo. Supponiamo di avere una torre in a4 (32) e di voler sapere se
   questa pu muovere, per esempio, di una casella alla sua sinistra:
   sottraiamo 1 al valore della sua casella (a4, valore 32), e otteniamo quindi
   31, che  il valore della casella h5. La torre, ovviamente, non pu muovere in h5,
   ma noi non possiamo saperlo senza fare una certa quantit di pesanti calcoli.
   Perci noi prendiamo il valore corrispondente alla casella a4 (dal vettore mailbox64), 
   ottenendo dunque 61. Quindi sottraiamo uno da questo valore (61-1=60) e andiamo a vedere
   cosa c' in mailbox[60]. In questo caso c' -1, perci la torre finisce fuori dalla 
   scacchiera e noi possiamo quindi rinunciare al proposito di spostarla di una casella 
   a sinistra...
   Puoi andare a vedere in che modo  usato il vettore mailbox[] guardando nella funzione
   attack() nel file board.c. */


int mailbox[120] = {
	 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	 -1,  0,  1,  2,  3,  4,  5,  6,  7, -1,
	 -1,  8,  9, 10, 11, 12, 13, 14, 15, -1,
	 -1, 16, 17, 18, 19, 20, 21, 22, 23, -1,
	 -1, 24, 25, 26, 27, 28, 29, 30, 31, -1,
	 -1, 32, 33, 34, 35, 36, 37, 38, 39, -1,
	 -1, 40, 41, 42, 43, 44, 45, 46, 47, -1,
	 -1, 48, 49, 50, 51, 52, 53, 54, 55, -1,
	 -1, 56, 57, 58, 59, 60, 61, 62, 63, -1,
	 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1
};

int mailbox64[64] = {
	21, 22, 23, 24, 25, 26, 27, 28,
	31, 32, 33, 34, 35, 36, 37, 38,
	41, 42, 43, 44, 45, 46, 47, 48,
	51, 52, 53, 54, 55, 56, 57, 58,
	61, 62, 63, 64, 65, 66, 67, 68,
	71, 72, 73, 74, 75, 76, 77, 78,
	81, 82, 83, 84, 85, 86, 87, 88,
	91, 92, 93, 94, 95, 96, 97, 98
};


/* slide, offsets, e offset sono, fondamentalmente, le direzioni in
   cui i pezzi possono muoversi. Se slide per un certo pezzo  FALSE, 
   questo pu muoversi solo di una casa alla volta in una qualche direzione. 
   Offsets  il numero delle direzioni in cui pu il pezzo muovere, e offset  
    il vettore contenente le direzioni effettive per quel pezzo. */

BOOL slide[6] = {
	FALSE, FALSE, TRUE, TRUE, TRUE, FALSE
};

int offsets[6] = {
	0, 8, 4, 4, 8, 8
};

int offset[6][8] = {
	{ 0, 0, 0, 0, 0, 0, 0, 0 },
	{ -21, -19, -12, -8, 8, 12, 19, 21 },
	{ -11, -9, 9, 11, 0, 0, 0, 0 },
	{ -10, -1, 1, 10, 0, 0, 0, 0 },
	{ -11, -10, -9, -1, 1, 9, 10, 11 },
	{ -11, -10, -9, -1, 1, 9, 10, 11 }
};


/* Questo  il vettore castle_mask. Noi possiamo usarlo per
   determinare i permessi per gli arrocchi dopo che una mossa
    stata giocata. Quello che dobbiamo fare  un And-logico
   tra i bits dell'arrocco con i bits del castle_mask per
   la casa interessata dall'ultima mossa effettuata.
   Supponiamo che castle sia uguale ad 1,
   intendendo con ci che il bianco pu ancora arroccare corto.
   Ora noi giochiamo una mossa che cattura la torre in h1 (per esempio ...Axh1[N.d.T]).
   Facciamo quindi un AND-logico fra il bit dell'arrocco corto per il bianco ed
   il castle_mask[63] (dove 63  il valore di h1 in mailbox [N.d.T]), 
   ottenendo 1&14, memorizzando il risultato in  castle, che diventa 0 (01110&00001=00000).
   In definitiva, una volta catturata la torre in h1, al bianco non  pi consentito
   arroccare corto. */

int castle_mask[64] = {
	 7, 15, 15, 15,  3, 15, 15, 11,
	15, 15, 15, 15, 15, 15, 15, 15,
	15, 15, 15, 15, 15, 15, 15, 15,
	15, 15, 15, 15, 15, 15, 15, 15,
	15, 15, 15, 15, 15, 15, 15, 15,
	15, 15, 15, 15, 15, 15, 15, 15,
	15, 15, 15, 15, 15, 15, 15, 15,
	13, 15, 15, 15, 12, 15, 15, 14
};


/* le lettere assgnate ai pezzi, per la print_board() :
   P	-->	(P)awn 		--> Pedone
   N    -->	k(N)ight	--> Cavallo
   B	-->	(B)ishop	--> Alfiere
   R	-->	(R)ook		--> Torre
   Q	-->	(Q)ueen		--> Regina o Donna
   K	-->	(K)ing		--> Re
*/
char piece_char[6] = {
	'P', 'N', 'B', 'R', 'Q', 'K'
};


/* la posizione iniziale della scacchiera */

int init_color[64] = {
	1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1,
	6, 6, 6, 6, 6, 6, 6, 6,
	6, 6, 6, 6, 6, 6, 6, 6,
	6, 6, 6, 6, 6, 6, 6, 6,
	6, 6, 6, 6, 6, 6, 6, 6,
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0
};

int init_piece[64] = {
	3, 1, 2, 4, 5, 2, 1, 3,
	0, 0, 0, 0, 0, 0, 0, 0,
	6, 6, 6, 6, 6, 6, 6, 6,
	6, 6, 6, 6, 6, 6, 6, 6,
	6, 6, 6, 6, 6, 6, 6, 6,
	6, 6, 6, 6, 6, 6, 6, 6,
	0, 0, 0, 0, 0, 0, 0, 0,
	3, 1, 2, 4, 5, 2, 1, 3
};
