/*
 *	DEFS.H
 *	Tom Kerrigan's Simple Chess Program (TSCP)
 *
 *	Copyright 1997 Tom Kerrigan
 *
 *	Traduzione in italiano di Raffaele Carnuccio
 */


#define BOOL			int
#define TRUE			1
#define FALSE			0

#define GEN_STACK		1120
#define MAX_PLY			32
#define HIST_STACK		400

#define LIGHT			0
#define DARK			1

#define PAWN			0
#define KNIGHT			1
#define BISHOP			2
#define ROOK			3
#define QUEEN			4
#define KING			5

#define EMPTY			6

/* case utili */
#define A1				56
#define B1				57
#define C1				58
#define D1				59
#define E1				60
#define F1				61
#define G1				62
#define H1				63
#define A8				0
#define B8				1
#define C8				2
#define D8				3
#define E8				4
#define F8				5
#define G8				6
#define H8				7

#define ROW(x)			(x >> 3)
#define COL(x)			(x & 7)


/* Questa  la descrizione base di una mossa.promote  il pezzo
   a cui si promuove se la mossa  la promozione di un pedone.
   bits  un campo di bits che descrive la mossa,
   con i seguenti bits:

   1	capture						 cattura  			 			[N.d.T.]	
   2	castle						 arrocco 		 				[N.d.T.]
   4	en passant capture			 presa en-passant 	 			[N.d.T.]	
   8	pushing a pawn 2 squares	 spinta del pedone di 2 case	[N.d.T.]
   16	pawn move					 mossa di pedone				[N.d.T.]
   32	promote						 promozione						[N.d.T.]	*/


/*   E' unito con un intero per poter comparare due mosse rapidamente 
     e facilmente. 
     Il compilatore infatti produce una comparazione fra due interi pi rapida 
     di quella che produrrebbe fra due strutture complete */

typedef struct {
	char from;
	char to;
	char promote;
	char bits;
} move_bytes;

typedef union {
	move_bytes b;
	int u;
} move;

/* un elemento del move stack. esso  composto da una mossa
   e da uno score, in modo da poter essere ordinato dalla
   funzione di ricerca. */
typedef struct {
	move m;
	int score;
} gen_t;

/* un elemento dell' history stack,con le informazioni
   necessarie per tornare indietro di una mossa. */
typedef struct {
	move m;
	int capture;
	int castle;
	int ep;
	int fifty;
} hist_t;
