/*
 *	EVAL.C
 *	Tom Kerrigan's Simple Chess Program (TSCP)
 *
 *	Copyright 1997 Tom Kerrigan
 *
 *	Traduzione in italiano di Raffaele Carnuccio
 */


#include <string.h>
#include "defs.h"
#include "data.h"
#include "protos.h"


#define DOUBLED_PAWN_PENALTY		10   /* penalit per i pedoni doppiati [N.d.T] */
#define ISOLATED_PAWN_PENALTY		20   /* penalit per i pedoni isolati [N.d.T] */ 
#define BACKWARDS_PAWN_PENALTY		8    /* penalit per i pedoni arretrati [N.d.T] */
#define PASSED_PAWN_BONUS			20   /* bonus per i pedoni passati [N.d.T] */
#define ROOK_SEMI_OPEN_FILE_BONUS	10   /* bonus per torre su colonna semiaperta [N.d.T] */
#define ROOK_OPEN_FILE_BONUS		15   /* bonus per torre su colonna aperta [N.d.T] */
#define ROOK_ON_SEVENTH_BONUS		20   /* bonus per torre in settima [N.d.T] */


/* il valore dei pezzi:
    	100 per il pedone, 
	300 per il cavallo
	300 per l'alfiere
	500 per la torre
	900 per la regina
	0 per il re  		[N.d.T.]*/

int piece_value[6] = {
	100, 300, 300, 500, 900, 0
};

/* I vettori "pcsq" costituiscono la tabella pezzo/casa. 
   Questi aggiungono una certa quantit al valore materiale di una pezzo
   a seconda di dove questo si trovi sulla scacchiera.
   (Ci favorisce, ad esempio, la centralizzazione dei pezzi o la spinta 
   dei pedoni centrali [N.d.T]) */

/* Tabella per il pedone */
int pawn_pcsq[64] = {
	  0,   0,   0,   0,   0,   0,   0,   0,
	  5,  10,  15,  20,  20,  15,  10,   5,
	  4,   8,  12,  16,  16,  12,   8,   4,
	  3,   6,   9,  12,  12,   9,   6,   3,
	  2,   4,   6,   8,   8,   6,   4,   2,
	  1,   2,   3, -10, -10,   3,   2,   1,
	  0,   0,   0, -40, -40,   0,   0,   0,
	  0,   0,   0,   0,   0,   0,   0,   0
};

/* Tabella per il cavallo */
int knight_pcsq[64] = {
	-10, -10, -10, -10, -10, -10, -10, -10,
	-10,   0,   0,   0,   0,   0,   0, -10,
	-10,   0,   5,   5,   5,   5,   0, -10,
	-10,   0,   5,  10,  10,   5,   0, -10,
	-10,   0,   5,  10,  10,   5,   0, -10,
	-10,   0,   5,   5,   5,   5,   0, -10,
	-10,   0,   0,   0,   0,   0,   0, -10,
	-10, -30, -10, -10, -10, -10, -30, -10
};

/* Tabella per l'alfiere */
int bishop_pcsq[64] = {
	-10, -10, -10, -10, -10, -10, -10, -10,
	-10,   0,   0,   0,   0,   0,   0, -10,
	-10,   0,   5,   5,   5,   5,   0, -10,
	-10,   0,   5,  10,  10,   5,   0, -10,
	-10,   0,   5,  10,  10,   5,   0, -10,
	-10,   0,   5,   5,   5,   5,   0, -10,
	-10,   0,   0,   0,   0,   0,   0, -10,
	-10, -10, -20, -10, -10, -20, -10, -10
};

/* Tabella per il re */
int king_pcsq[64] = {
	-40, -40, -40, -40, -40, -40, -40, -40,
	-40, -40, -40, -40, -40, -40, -40, -40,
	-40, -40, -40, -40, -40, -40, -40, -40,
	-40, -40, -40, -40, -40, -40, -40, -40,
	-40, -40, -40, -40, -40, -40, -40, -40,
	-40, -40, -40, -40, -40, -40, -40, -40,
	-20, -20, -20, -20, -20, -20, -20, -20,
	  0,  20,  40, -20,   0, -20,  40,  20
};

/* Tabella per il re (nel finale di partita) */
int king_endgame_pcsq[64] = {
	  0,  10,  20,  30,  30,  20,  10,   0,
	 10,  20,  30,  40,  40,  30,  20,  10,
	 20,  30,  40,  50,  50,  40,  30,  20,
	 30,  40,  50,  60,  60,  50,  40,  30,
	 30,  40,  50,  60,  60,  50,  40,  30,
	 20,  30,  40,  50,  50,  40,  30,  20,
	 10,  20,  30,  40,  40,  30,  20,  10,
	  0,  10,  20,  30,  30,  20,  10,   0
};

/* Il vettore Flip  usato per calcolare il valore pezzo casa
   per i pezzi (DARK). 
   Il valore pezzo/casa di un pedone LIGHT  pawn_pcsq[sq] mentre per un pedone
   DARK  pawn_pcsq[flip[sq]] */

int flip[64] = {
	 56,  57,  58,  59,  60,  61,  62,  63,
	 48,  49,  50,  51,  52,  53,  54,  55,
	 40,  41,  42,  43,  44,  45,  46,  47,
	 32,  33,  34,  35,  36,  37,  38,  39,
	 24,  25,  26,  27,  28,  29,  30,  31,
	 16,  17,  18,  19,  20,  21,  22,  23,
	  8,   9,  10,  11,  12,  13,  14,  15,
	  0,   1,   2,   3,   4,   5,   6,   7
};

/* pawn_rank[x][y]  la traversa del pedone meno avanzato di colore x sulla
   colonna y - 1. Ci sono "buffer files" sulla destra e sulla sinistra 
   per evitare speciali casi-logici pi avanti. 
   Se non ci sono pedoni sulla traversa, assumiamo che ce ne sia uno fuori
   dalla scacchiera (0 per LIGHT e 7 per DARK). Questo rende facile
   testare la presenza di pedoni su di una colonna e semplica alcune valutazioni 
   pedonali. */

int pawn_rank[2][10];

int piece_mat[2];  /* il valore dei pezzi dei due giocatori */
int pawn_mat[2];   /* il valore dei pedoni dei due giocatori */

int eval()
{
	int i;
	int f;  /* colonna */
	int score[2];  /* il punteggio complessivo dei due giocatori */

	/* questo  il primo passo: settare pawn_rank, piece_mat, e pawn_mat. */
	for (i = 0; i < 10; ++i) {
		pawn_rank[LIGHT][i] = 0;
		pawn_rank[DARK][i] = 7;
	}
	piece_mat[LIGHT] = 0;
	piece_mat[DARK] = 0;
	pawn_mat[LIGHT] = 0;
	pawn_mat[DARK] = 0;
	for (i = 0; i < 64; ++i) {
		if (color[i] == EMPTY)
			continue;
		if (piece[i] == PAWN) {
			pawn_mat[color[i]] += piece_value[PAWN];
			f = COL(i) + 1;  /* aggiungo 1 per la colonna extra dell'array */
			if (color[i] == LIGHT) {
				if (pawn_rank[LIGHT][f] < ROW(i))
					pawn_rank[LIGHT][f] = ROW(i);
			}
			else {
				if (pawn_rank[DARK][f] > ROW(i))
					pawn_rank[DARK][f] = ROW(i);
			}
		}
		else
			piece_mat[color[i]] += piece_value[piece[i]];
	}

	/* questo  il secondo passo: valutare ogni pezzo */
	score[LIGHT] = piece_mat[LIGHT] + pawn_mat[LIGHT];
	score[DARK] = piece_mat[DARK] + pawn_mat[DARK];
	for (i = 0; i < 64; ++i) {
		if (color[i] == EMPTY)
			continue;
		if (color[i] == LIGHT) {
			switch (piece[i]) {
				case PAWN:
					score[LIGHT] += eval_light_pawn(i);
					break;
				case KNIGHT:
					score[LIGHT] += knight_pcsq[i];
					break;
				case BISHOP:
					score[LIGHT] += bishop_pcsq[i];
					break;
				case ROOK:
					if (pawn_rank[LIGHT][COL(i) + 1] == 0) {
						if (pawn_rank[DARK][COL(i) + 1] == 7)
							score[LIGHT] += ROOK_OPEN_FILE_BONUS;
						else
							score[LIGHT] += ROOK_SEMI_OPEN_FILE_BONUS;
					}
					if (ROW(i) == 1)
						score[LIGHT] += ROOK_ON_SEVENTH_BONUS;
					break;
				case KING:
					if (piece_mat[DARK] <= 1200)
						score[LIGHT] += king_endgame_pcsq[i];
					else
						score[LIGHT] += eval_light_king(i);
					break;
			}
		}
		else {
			switch (piece[i]) {
				case PAWN:
					score[DARK] += eval_dark_pawn(i);
					break;
				case KNIGHT:
					score[DARK] += knight_pcsq[flip[i]];
					break;
				case BISHOP:
					score[DARK] += bishop_pcsq[flip[i]];
					break;
				case ROOK:
					if (pawn_rank[DARK][COL(i) + 1] == 7) {
						if (pawn_rank[LIGHT][COL(i) + 1] == 0)
							score[DARK] += ROOK_OPEN_FILE_BONUS;
						else
							score[DARK] += ROOK_SEMI_OPEN_FILE_BONUS;
					}
					if (ROW(i) == 6)
						score[DARK] += ROOK_ON_SEVENTH_BONUS;
					break;
				case KING:
					if (piece_mat[LIGHT] <= 1200)
						score[DARK] += king_endgame_pcsq[flip[i]];
					else
						score[DARK] += eval_dark_king(i);
					break;
			}
		}
	}

	/* il vettore score[]  settato, ora restituisce il punteggio relativo
	   al lato cui tocca muovere */
	if (side == LIGHT)
		return score[LIGHT] - score[DARK];
	return score[DARK] - score[LIGHT];
}

int eval_light_pawn(int sq)
{
	int r;  /* il valore di ritorno */
	int f;  /* la colonna del pedone */

	r = 0;
	f = COL(sq) + 1;

	r += pawn_pcsq[sq];

	/* Se c' un pedone dietro a questo, allora  doppiato */
	if (pawn_rank[LIGHT][f] > ROW(sq))
		r -= DOUBLED_PAWN_PENALTY;

	/* se non ci sono pedoni amici sulle colonne adiacenti a 
	   questo pedone, allora esso  isolato */
	if ((pawn_rank[LIGHT][f - 1] == 0) &&
			(pawn_rank[LIGHT][f + 1] == 0))
		r -= ISOLATED_PAWN_PENALTY;

	/* se il pedone non  isolato, potrebbe essere arretrato */
	else if ((pawn_rank[LIGHT][f - 1] < ROW(sq)) &&
			(pawn_rank[LIGHT][f + 1] < ROW(sq)))
		r -= BACKWARDS_PAWN_PENALTY;

	/* aggiunge un bonus se il pedone  passato */
	if ((pawn_rank[DARK][f - 1] >= ROW(sq)) &&
			(pawn_rank[DARK][f] >= ROW(sq)) &&
			(pawn_rank[DARK][f + 1] >= ROW(sq)))
		r += (7 - ROW(sq)) * PASSED_PAWN_BONUS;

	return r;
}

int eval_dark_pawn(int sq)
{
	int r;  /* il valore di ritorno */
	int f;  /* la colonna del pedone */

	r = 0;
	f = COL(sq) + 1;

	r += pawn_pcsq[flip[sq]];

	/* Se c' un pedone dietro a questo, allora  doppiato */
	if (pawn_rank[DARK][f] < ROW(sq))
		r -= DOUBLED_PAWN_PENALTY;

	/* se non ci sono pedoni amici sulle colonne adiacenti a 
	   questo pedone, allora esso  isolato */
	if ((pawn_rank[DARK][f - 1] == 7) &&
			(pawn_rank[DARK][f + 1] == 7))
		r -= ISOLATED_PAWN_PENALTY;

	/* se il pedone non  isolato, potrebbe essere arretrato */
	else if ((pawn_rank[DARK][f - 1] > ROW(sq)) &&
			(pawn_rank[DARK][f + 1] > ROW(sq)))
		r -= BACKWARDS_PAWN_PENALTY;

	/* aggiunge un bonus se il pedone  passato */
	if ((pawn_rank[LIGHT][f - 1] <= ROW(sq)) &&
			(pawn_rank[LIGHT][f] <= ROW(sq)) &&
			(pawn_rank[LIGHT][f + 1] <= ROW(sq)))
		r += ROW(sq) * PASSED_PAWN_BONUS;

	return r;
}

int eval_light_king(int sq)
{
	int r;  /* il valore di ritorno */
	int i;

	r = king_pcsq[sq];

	/* se il re  arroccato, usa una speciale funzione per valutare i
	   pedoni secondo il lato di appartenenza */
	if (COL(sq) < 3) {
		r += eval_lkp(1);
		r += eval_lkp(2);
		r += eval_lkp(3) / 2;  /* i pedoni delle colonne di alfiere sono
					  valutati meno severamente  */
	}
	else if (COL(sq) > 4) {
		r += eval_lkp(8);
		r += eval_lkp(7);
		r += eval_lkp(6) / 2;
	}

	/* altrimenti, stima una penalit da applicare nel caso in cui
	   vi siano colonne aperte nei pressi del re */
	else {
		for (i = COL(sq); i <= COL(sq) + 2; ++i)
			if ((pawn_rank[LIGHT][i] == 0) &&
					(pawn_rank[DARK][i] == 7))
				r -= 10;
	}

	/* la sicurezza del re  valutata in accordo al materiale dell'avversario;
	   la premessa  che se il tuo avversario ha abbastanza pezzi per attaccarti, 
           la sicurezza del tuo re non pu comunque essere buona */
	r *= piece_mat[DARK];
	r /= 3100;

	return r;
}

/* eval_lkp(f) valuta il pedone light sulla colonna f */

int eval_lkp(int f)
{
	int r = 0;

	if (pawn_rank[LIGHT][f] == 6);  /* il pedone non ha mosso */
	else if (pawn_rank[LIGHT][f] == 5)
		r -= 10;  /* il pedone ha mosso di una casa */
	else if (pawn_rank[LIGHT][f] != 0)
		r -= 20;  /* il pedone ha mosso per + di una casa */
	else
		r -= 25;  /* non ci sono pedoni sulla colonna */

	if (pawn_rank[DARK][f] == 7)
		r -= 15;  /* non ci sono pedoni nemici */
	else if (pawn_rank[DARK][f] == 5)
		r -= 10;  /* pedone nemico sulla terza traversa */
	else if (pawn_rank[DARK][f] == 4)
		r -= 5;   /* pedone nemico sulla quarta traversa */

	return r;
}

int eval_dark_king(int sq)
{
	int r;
	int i;

	r = king_pcsq[flip[sq]];
	if (COL(sq) < 3) {
		r += eval_dkp(1);
		r += eval_dkp(2);
		r += eval_dkp(3) / 2;
	}
	else if (COL(sq) > 4) {
		r += eval_dkp(8);
		r += eval_dkp(7);
		r += eval_dkp(6) / 2;
	}
	else {
		for (i = COL(sq); i <= COL(sq) + 2; ++i)
			if ((pawn_rank[LIGHT][i] == 0) &&
					(pawn_rank[DARK][i] == 7))
				r -= 10;
	}
	r *= piece_mat[LIGHT];
	r /= 3100;
	return r;
}

int eval_dkp(int f)
{
	int r = 0;

	if (pawn_rank[DARK][f] == 1);
	else if (pawn_rank[DARK][f] == 2)
		r -= 10;
	else if (pawn_rank[DARK][f] != 7)
		r -= 20;
	else
		r -= 25;

	if (pawn_rank[LIGHT][f] == 0)
		r -= 15;
	else if (pawn_rank[LIGHT][f] == 2)
		r -= 10;
	else if (pawn_rank[LIGHT][f] == 3)
		r -= 5;

	return r;
}
