/*
 *	SEARCH.C
 *	Tom Kerrigan's Simple Chess Program (TSCP)
 *
 *	Copyright 1997 Tom Kerrigan
 *
 *	Traduzione in italiano di Raffaele Carnuccio
 */


#include <stdio.h>
#include <string.h>
#include "defs.h"
#include "data.h"
#include "protos.h"


/* guarda all'inizio di think() */
#include <setjmp.h>
jmp_buf env;
BOOL stop_search;


/* think() chiama search() iterativamente. Le statistiche
   di ricerca stampate a video dipendono dal valore di output:
   0 = nessun output
   1 = output normale
   2 = output formato xboard */

void think(int output)
{
	int i, j, x;

	/* un p di codice che ci permette di ritornare qui da think() 
           quando  ora */
	stop_search = FALSE;
	setjmp(env);
	if (stop_search) {
		
		/* rende certo che si torni indietro sulla linea che stamo analizzando */
		while (ply)
			takeback();
		return;
	}

	start_time = get_ms();
	stop_time = start_time + max_time;

	ply = 0;
	nodes = 0;
	memset(pv, 0, sizeof(pv));
	memset(history, 0, sizeof(history));
	if (output == 1)
		printf("ply      nodes  score  pv\n");
	for (i = 1; i <= max_depth; ++i) {
		follow_pv = TRUE;
		x = search(-10000, 10000, i);
		if (output == 1)
			printf("%3d  %9d  %5d ", i, nodes, x);
		else if (output == 2)
			printf("%d %d %d %d",
					i, x, (get_ms() - start_time) / 10, nodes);
		if (output) {
			for (j = 0; j < pv_length[0]; ++j)
				printf(" %s", move_str(pv[0][j].b));
			printf("\n");
			fflush(stdout);
		}
		if (x > 9000 || x < -9000)
			break;
	}
}


/* search() f giusto questo, in modo negamax */

int search(int alpha, int beta, int depth)
{
	int i, j, x;
	BOOL c, f;

	/* siamo alla profondit in cui volevamo essere; 
	   chiama quiesce() per ottenere un punteggio ragionevole
	   e restituirlo. */
	if (!depth)
		return quiesce(alpha,beta);
	++nodes;

	/* chiama checkup() ogni 1024 nodi */
	if ((nodes & 1023) == 0)
		checkup();

	pv_length[ply] = ply;

	/* se questa non  la radice dell' albero di ricerca (dove dobbiamo selezionare 
           una mossa e non possiamo semplicemente restituire  0), va a vedere che la 
	   posizione non sia una ripetizione.
	   Se  cos,  suppone che questa linea di gioco sia patta e restituisce 0. */

	if (ply && reps())
		return 0;

	/* are we too deep? */
	if (ply >= MAX_PLY - 1)
		return eval();
	if (hply >= HIST_STACK - 1)
		return eval();

	/* siamo sotto scacco? se si, cerchiamo pi in profondit */
	c = in_check(side);
	if (c)
		++depth;
	gen();
	if (follow_pv)  /* stiamo seguendo la variante principale ?*/
		sort_pv();
	f = FALSE;

	/* looppa fra le mosse */
	for (i = first_move[ply]; i < first_move[ply + 1]; ++i) {
		sort(i);
		if (!makemove(gen_dat[i].m.b))
			continue;
		f = TRUE;
		x = -search(-beta, -alpha, depth - 1);
		takeback();
		if (x > alpha) {

			/* questa mossa ha causato un taglio, quindi si incrementa il valore
			   di history cos viene ordinata in alto la prossima volta che la 
			   cerchiamo */
			history[(int)gen_dat[i].m.b.from][(int)gen_dat[i].m.b.to] += depth;
			if (x >= beta)
				return beta;
			alpha = x;

			/* aggiorna la variante principale */
			pv[ply][ply] = gen_dat[i].m;
			for (j = ply + 1; j < pv_length[ply + 1]; ++j)
				pv[ply][j] = pv[ply + 1][j];
			pv_length[ply] = pv_length[ply + 1];
		}
	}

	/* non ci sono mosse legali? allora c' un matto o uno stallo */
	if (!f) {
		if (c)
			return -10000 + ply;
		else
			return 0;
	}

	/* patta per la regola delle 50 mosse */
	if (fifty >= 100)
		return 0;
	return alpha;
}


/*  
   quiesce()  una funzione ricorsiva di ricerca del minimax con i tagli alfa-beta. 
   In altre parole, negamax. 
   Essa fondamentalmente cerca solo le sequenze di catture  permette alla funzione
   di valutazione di tagliare la ricerca (e settare alpha).
   L'idea  di trovare una posizione dove non ci sia da andare troppo lontano [con 
   scacchi e scambi tattici? N.d.T.] in modo che la funzione statica di valutazione
   possa funzionare correttamente. */

/* Il fatto che la funzione di valutazione possa tagliare la ricerca, significa che
   non siamo obbligati ad eseguire mosse di cattura. Se la valutazione statica della
   posizione  abbastanza buona, significa che siamo usciti dagli scambi tattici con
   una buona posizione, e potremo probabilmente effettuare una mossa "tranquilla" che
   non implichi ulteriori scambi di pezzi. 

   [Questo e altri commenti qui presenti, non provengono dal codice originale di 
   Tom Kerrigan, ma sono stati aggiunti da Jose Carlos Martinez Galan, traduttore 
   del codice di TSCP in castigliano, e ritradotte in italiano dal sottoscritto.
   Vedi file Readme.txt allegato al presente codice N.d.T]		*/


int quiesce(int alpha,int beta)
{
	int i, j, x;

	++nodes;

	/* chiama checkup() ogni 1024 nodi */
	if ((nodes & 1023) == 0)
		checkup();

	pv_length[ply] = ply;

	/* siamo troppo in profondit? */
	if (ply >= MAX_PLY - 1)
		return eval();
	if (hply >= HIST_STACK - 1)
		return eval();

	/* controlla con la funzione di valutazione */
	x = eval();
	if (x >= beta)
		return beta;
	if (x > alpha)
		alpha = x;

	gen_caps();
	if (follow_pv)  /* stiamo seguendo la variante principale? */
		sort_pv();

	/* looppa tra le mosse */
	for (i = first_move[ply]; i < first_move[ply + 1]; ++i) {
		sort(i);
		if (!makemove(gen_dat[i].m.b))
			continue;
		x = -quiesce(-beta, -alpha);
		takeback();
		if (x > alpha) {
			if (x >= beta)
				return beta;
			alpha = x;

			/* aggiorna la variante principale */
			pv[ply][ply] = gen_dat[i].m;
			for (j = ply + 1; j < pv_length[ply + 1]; ++j)
				pv[ply][j] = pv[ply + 1][j];
			pv_length[ply] = pv_length[ply + 1];
		}
	}
	return alpha;
}


/* reps() restituisce il numero di volte in cui la posizione 
    stata ripetuta. 
   Ringrazio John Stanback per questo intelligente algoritmo. */

int reps()
{
	int i;
	int b[64];
	int c = 0;  /* conta le case che hanno valori differenti
		       dalla posizione corrente */
	int r = 0;  /* numero di ripetizioni */

	/*  una ripetizione impossibile? */
	if (fifty <= 3)
		return 0;

	memset(b, 0, sizeof(b));

	/* loppa tra le mosse reversibili */
	for (i = hply - 1; i >= hply - fifty - 1; --i) {
		if (++b[hist_dat[i].m.b.from] == 0)
			--c;
		else
			++c;
		if (--b[hist_dat[i].m.b.to] == 0)
			--c;
		else
			++c;
		if (c == 0)
			++r;
	}

	return r;
}


/* sort_pv()  chiamata quando la funzione di ricerca sta seguendo
   la PV (Variante Principale). Essa guarda nella lista di mosse dell'
   attuale profondit per vedere se la mossa PV  qui. 
   Se si, aggiunge 10,000,000 al punteggio della mossa cos essa
    giocata prima dalla funzione di ricerca.
   Se no, follow_pv resta FALSE e search() si ferma chiamando sort_pv(). */

void sort_pv()
{
	int i;

	follow_pv = FALSE;
	for(i = first_move[ply]; i < first_move[ply + 1]; ++i)
		if (gen_dat[i].m.u == pv[0][ply].u) {
			follow_pv = TRUE;
			gen_dat[i].score += 10000000;
			return;
		}
}


/* sort() cerca la lista di mosse della profondit corrente da 'from'
   per trovare le mossa con il punteggio pi alto. Poi scambia
   tale mossa con la mossa 'from', per tenere in cima alla lista la
   mossa migliore che (si spera) produrr un taglio. */

void sort(int from)
{
	int i;
	int bs;  /* migliore score */
	int bi;  /* migliore i */
	gen_t g;

	bs = -1;
	bi = from;
	for (i = from; i < first_move[ply + 1]; ++i)
		if (gen_dat[i].score > bs) {
			bs = gen_dat[i].score;
			bi = i;
		}
	g = gen_dat[from];
	gen_dat[from] = gen_dat[bi];
	gen_dat[bi] = g;
}


/* checkup()  chiamato occasionalmente durante la ricerca. */

void checkup()
{
	/*  il tempo di time up dell'engine?
          if  cos, longjmp torna all'inizio di think() */
	if (get_ms() >= stop_time) {
		stop_search = TRUE;
		longjmp(env, 0);
	}
}
